module VGA_GraphicsController(CLK, nRESET,
                              IO_REG_SEL, IO_WRITE, IO_READ, IO_DATA_IN, IO_DATA_OUT, IO_RDY,
                              PRAM_DI, PRAM_DO, PRAM_ADDR, PRAM_WR, PRAM_RD, PRAM_RDY);

input CLK;
input nRESET;

/* I/O Interface */
input            IO_WRITE;      // write CMD or parameter Data to VGA controller
input            IO_READ;       // read status or data from VGA controller
input      [3:0] IO_REG_SEL;    // Register select (0=CMD, 1=Data1, ..., 15=Data15)
input      [7:0] IO_DATA_IN;    // CMD/Data sent to VGA controller
output reg [7:0] IO_DATA_OUT;   // Status or data read from VGA controller
output           IO_RDY;        // I/O operation has been finished

/* VGA pixel RAM interface */
input      [15:0] PRAM_DI;      // data from Pixel RAM
output reg [15:0] PRAM_DO;      // data to Pixel RAM
output reg [21:0] PRAM_ADDR;
output reg        PRAM_WR;
output reg        PRAM_RD;
input             PRAM_RDY;

wire IO_Access;
assign IO_Access = IO_READ || IO_WRITE;
assign IO_RDY = Status_Ready;

reg Status_Busy;    // command triggered and executing, waiting for completion
reg Status_Ready;   // command executed, resulting data are available if there are any
reg Status_Waiting; // waiting for further parameter data for a triggered command
reg Status_Error;   // command execution error, error code in register DATA1

wire [7:0] Status;

assign Status = { 4'b0 , Status_Error, Status_Waiting, Status_Ready, Status_Busy };

localparam VGA_REG_CMD          = 4'd0;

localparam VGA_ERR_WRONG_DIM    = 8'd3;     // wrong dimensions

localparam VGA_RESET            = 8'd0;
localparam VGA_SET_X            = 8'd2;     // set cursor x position
localparam VGA_DRAW_LINE        = 8'd21;    // draw a line from cusor (x0,y0) to a given x1,y1

localparam VGA_OP_PAT           = 8'd0;
localparam VGA_OP_INV           = 8'd1;
localparam VGA_OP_CLR           = 8'd2;
localparam VGA_OP_SET           = 8'd3;

localparam VGA_DEFAULT_YSIZE    = 16'd400;
localparam VGA_WPL              = 16'd512;

localparam IDLE                 = 8'd0;
localparam INIT                 = 8'd1;
localparam READY                = 8'd4;
localparam DRAW_LINE_1          = 8'd19;
localparam DRAW_LINE_2          = 8'd20;
localparam DRAW_LINE_3          = 8'd21;
localparam DRAW_LINE_4          = 8'd22;

reg  [7:0]  State = IDLE;
reg  [7:0]  Color;

reg  [15:0] DX0;
reg  [15:0] DY0;
reg  [15:0] DX1;
reg  [15:0] DY1;

reg  DrawLine_Start = 1'b0;
reg  DrawLine_Output = 1'b0;
wire DrawLine_Drawing;
wire DrawLine_Complete;
wire DrawLine_Done;
wire [15:0] DrawLine_X;
wire [15:0] DrawLine_Y;

wire [1:0] Pattern[0:1];
assign Pattern[0] = 2'b01;
assign Pattern[1] = 2'b10;

reg [7:0] Data[0:15];

always @(posedge CLK)
    if (!nRESET)
        State <= INIT;
    else
        case (State)
            INIT:           begin
                                Status_Ready <= 1'b1;
                                Status_Busy <= 1'b0;
                                Status_Waiting <= 1'b0;
                                Status_Error <= 1'b0;
                                DX0 <= 0;
                                DY0 <= 0;
                                Color <= { 4'b0000, 4'b1111 };
                                DrawLine_Start <= 1'b0;
                                DrawLine_Output <= 1'b0;
                                State <= READY;
                            end

            IDLE:           if (~IO_Access)
                                Status_Busy <= 1'b0;
                            else begin
                                if (IO_READ) begin
                                    IO_DATA_OUT <= (IO_REG_SEL == VGA_REG_CMD) ? Status : Data[IO_REG_SEL];
                                    State <= READY;

                                end else if (IO_WRITE & IO_REG_SEL == VGA_REG_CMD) begin
                                    Status_Busy <= 1'b1;
                                    Status_Ready <= 1'b0;
                                    Status_Error <= 1'b0;

                                    case (IO_DATA_IN)
                                        VGA_RESET:      State <= INIT;

                                        VGA_SET_X:      begin
                                                            DX0 <= { Data[2], Data[1] };
                                                            State <= READY;
                                                        end
                                        // ...

                                        VGA_DRAW_LINE:  begin
                                                            DX1 <= { Data[3], Data[2] };
                                                            DY1 <= { Data[5], Data[4] };
                                                            DrawLine_Start <= 1'b1;
                                                            State <= DRAW_LINE_1;
                                                        end
                                        // ...

                                        default:        begin
                                                            State <= READY;
                                                            Status_Error <= 1'b1;
                                                            Data[1] <= VGA_ERR_WRONG_DIM;
                                                        end
                                    endcase;

                                end else if (IO_WRITE) begin
                                    Data[IO_REG_SEL] <= IO_DATA_IN;
                                    State <= READY;
                                end

                                else
                                    State <= IDLE;
                            end

            DRAW_LINE_1:    begin
                                DrawLine_Start <= 1'b0;
                                DrawLine_Output <= 1'b1;
                                State <= DRAW_LINE_2;
                            end
            DRAW_LINE_2:    if (DrawLine_Drawing) begin
                                DrawLine_Output <= 1'b0;
                                RAM_Addr <= DrawLine_Y * VGA_WPL + DrawLine_X[15:1];
                                RAM_Pattern <= Pattern[DrawLine_X[0]];
                                RAM_OpCode <= Data[1];
                                RAM_Mode <= RAM_ReadWrite;
                                RAM_Start <= 1'b1;
                                State <= DRAW_LINE_3;
                            end
            DRAW_LINE_3:    begin
                                RAM_Start <= 1'b0;
                                State <= RAM_Ready ? DRAW_LINE_4 : DRAW_LINE_3;
                            end
            DRAW_LINE_4:    State <= DrawLine_Complete ? READY : DRAW_LINE_1;

            // ...

            READY:          begin
                                Status_Ready <= 1'b1;
                                State <= (IO_Access) ? READY : IDLE;
                            end

            default:        State <= IDLE;
        endcase

localparam RAM_OnlyWrite = 1'd0;
localparam RAM_ReadWrite = 1'd1;

localparam RAM_IDLE     = 4'd0;
localparam RAM_START    = 4'd1;
localparam RAM_READY    = 4'd2;
localparam RAM_READ_1   = 4'd3;
localparam RAM_READ_2   = 4'd4;
localparam RAM_READ_3   = 4'd5;
localparam RAM_READ_4   = 4'd6;
localparam RAM_WRITE_1  = 4'd7;
localparam RAM_WRITE_2  = 4'd8;
localparam RAM_WRITE_3  = 4'd9;
localparam RAM_WRITE_4  = 4'd10;

reg  [3:0]  RAM_State;

reg         RAM_Ready;       // ready to start RWC (Read/Write Cycle)
reg         RAM_Start;       // start RWC
reg         RAM_Mode;        // RWC mode (0 - only write, 1 - read then write)
reg  [21:0] RAM_Addr;        //
reg  [3:0]  RAM_Pattern;
reg  [7:0]  RAM_OpCode;

reg  [15:0] pram_di;
wire [15:0] cpattern = { Color[7] & RAM_Pattern[1],
                         Color[6] & RAM_Pattern[1],
                         Color[5] & RAM_Pattern[1],
                         Color[4] & RAM_Pattern[1],
                         Color[3] & RAM_Pattern[1],
                         Color[2] & RAM_Pattern[1],
                         Color[1] & RAM_Pattern[1],
                         Color[0] & RAM_Pattern[1],
                         Color[7] & RAM_Pattern[0],
                         Color[6] & RAM_Pattern[0],
                         Color[5] & RAM_Pattern[0],
                         Color[4] & RAM_Pattern[0],
                         Color[3] & RAM_Pattern[0],
                         Color[2] & RAM_Pattern[0],
                         Color[1] & RAM_Pattern[0],
                         Color[0] & RAM_Pattern[0] };                         

always @(posedge CLK)
    if (!nRESET)
        begin
            RAM_Ready <= 1'b0;
            PRAM_WR <= 0;
            PRAM_RD <= 0;
        end
    else
        case (RAM_State)
            RAM_IDLE:       begin
                                RAM_Ready <= 1'b0;
                                if (RAM_Start) begin
                                    RAM_State <= RAM_START;
                                end
                            end
            RAM_START:      begin                                    
                                    PRAM_ADDR <= RAM_Addr;
                                    if (RAM_Mode == RAM_ReadWrite)
                                        RAM_State <= RAM_READ_1;
                                    else begin
                                        pram_di <= 16'b0;
                                        RAM_State <= RAM_WRITE_1;
                                    end
                            end

            RAM_READ_1:     begin
                                PRAM_RD <= 1;
                                RAM_State <= RAM_READ_2;
                            end
            RAM_READ_2:     RAM_State <= RAM_READ_3;
            RAM_READ_3:     RAM_State <= (PRAM_RDY) ? RAM_READ_4 : RAM_READ_3;
            RAM_READ_4:     begin
                                pram_di <= PRAM_DI;
                                PRAM_RD <= 0;
                                RAM_State <= RAM_WRITE_1;
                            end

            RAM_WRITE_1:    begin
                                case (RAM_OpCode)
                                    VGA_OP_PAT, // TODO
                                    VGA_OP_SET: PRAM_DO <= pram_di | cpattern;
                                    VGA_OP_INV: PRAM_DO <= pram_di ^ cpattern;
                                    VGA_OP_CLR: PRAM_DO <= pram_di & ~cpattern;
                                endcase
                                RAM_State <= RAM_WRITE_2;
                            end
            RAM_WRITE_2:    begin
                                PRAM_WR <= 1;
                                RAM_State <= RAM_WRITE_3;
                            end
            RAM_WRITE_3:    RAM_State <= RAM_WRITE_4;
            RAM_WRITE_4:    RAM_State <= (PRAM_RDY) ? RAM_READY : RAM_WRITE_4;

            RAM_READY:      begin
                                PRAM_RD <= 0;
                                PRAM_WR <= 0;
                                RAM_Ready <= 1'b1;
                                RAM_State <= RAM_IDLE;
                            end
        endcase

draw_line _VGA_DrawLine(
    .clk(CLK),
    .rst(~nRESET),
    .start(DrawLine_Start),
    .oe(DrawLine_Output),
    .x0(DX0),
    .y0(DY0),
    .x1(DX1),
    .y1(DY1),
    .x(DrawLine_X),
    .y(DrawLine_Y),
    .drawing(DrawLine_Drawing),
    .complete(DrawLine_Complete),
    .done(DrawLine_Done));
    defparam _VGA_DrawLine.CORDW = 16;

endmodule
